#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <cuda.h>

/* do a greyscale image by averaging the colors at each point */
void cpuGreyscale(const int nrows, const int ncols, 
               const int *in_r, const int *in_g, const int *in_b,
               int *out_r, int *out_g, int *out_b) {

    int pix = 0;
    for (pix = 0; pix < nrows*ncols; pix++) {
        int avg = (in_r[pix] + in_g[pix] + in_b[pix])/3;
        out_r[pix] = avg;
        out_g[pix] = avg;
        out_b[pix] = avg;
    }
    return;
}


#define CHK_CUDA(e) {if (e != cudaSuccess) {fprintf(stderr,"Error: %s\n", cudaGetErrorString(e)); exit(-1);}}
#define CHK_ERROR   {cudaThreadSynchronize(); cudaError_t e = cudaGetLastError(); if(e != cudaSuccess) {fprintf(stderr,"Error: %s\n", cudaGetErrorString(e)); exit(-1);}}


/* do a greyscale image by averaging the colors at a point */
__global__ void greyscaleKernel(const int nrows, const int ncols, 
                             const int *in_r_d, const int *in_g_d, const int *in_b_d,
                             int *out_r_d, int *out_g_d, int *out_b_d) {

        int i = threadIdx.x;

        if (i < nrows*ncols) {
            int avg = ( in_r_d[i] + in_g_d[i] + in_b_d[i] )/3 ;
            out_r_d[i] = avg;
            out_g_d[i] = avg;
            out_b_d[i] = avg;
        }

        return;
}

/* launch the kernel on our input image */
void gpuGreyscale(const int nrows, const int ncols, 
               const int *in_r, const int *in_g, const int *in_b,
               int *out_r, int *out_g, int *out_b) {

    int *in_r_d, *in_g_d, *in_b_d;
    int *out_r_d, *out_g_d, *out_b_d;

    CHK_CUDA( cudaMalloc(&in_r_d,  nrows*ncols*sizeof(int)) );
    CHK_CUDA( cudaMalloc(&in_g_d,  nrows*ncols*sizeof(int)) );
    CHK_CUDA( cudaMalloc(&in_b_d,  nrows*ncols*sizeof(int)) );

    CHK_CUDA( cudaMalloc(&out_r_d,  nrows*ncols*sizeof(int)) );
    CHK_CUDA( cudaMalloc(&out_g_d,  nrows*ncols*sizeof(int)) );
    CHK_CUDA( cudaMalloc(&out_b_d,  nrows*ncols*sizeof(int)) );

    CHK_CUDA( cudaMemcpy(in_r_d, in_r, nrows*ncols*sizeof(int), cudaMemcpyHostToDevice) );
    CHK_CUDA( cudaMemcpy(in_g_d, in_g, nrows*ncols*sizeof(int), cudaMemcpyHostToDevice) );
    CHK_CUDA( cudaMemcpy(in_b_d, in_b, nrows*ncols*sizeof(int), cudaMemcpyHostToDevice) );

    greyscaleKernel<<<1,nrows*ncols>>>(nrows, ncols, in_r_d, in_g_d, in_b_d, out_r_d, out_g_d, out_b_d);

    CHK_ERROR ;

    CHK_CUDA( cudaMemcpy(out_r, out_r_d, nrows*ncols*sizeof(int), cudaMemcpyDeviceToHost) );
    CHK_CUDA( cudaMemcpy(out_g, out_g_d, nrows*ncols*sizeof(int), cudaMemcpyDeviceToHost) );
    CHK_CUDA( cudaMemcpy(out_b, out_b_d, nrows*ncols*sizeof(int), cudaMemcpyDeviceToHost) );

    CHK_CUDA( cudaFree(in_r_d) );
    CHK_CUDA( cudaFree(in_g_d) );
    CHK_CUDA( cudaFree(in_b_d) );

    CHK_CUDA( cudaFree(out_r_d) );
    CHK_CUDA( cudaFree(out_g_d) );
    CHK_CUDA( cudaFree(out_b_d) ); 
    return;
}

/* Do a photographic negative by inverting the colours at each point */
void cpuInvert(const int nrows, const int ncols, 
               const int *in_r, const int *in_g, const int *in_b,
               int *out_r, int *out_g, int *out_b) {

    int row, col;
    int pix = 0;
    for (row = 0; row<nrows; row++) {
        for (col = 0; col<ncols; col++) {
            out_r[pix] = 255 - in_r[pix];
            out_g[pix] = 255 - in_g[pix];
            out_b[pix] = 255 - in_b[pix];
            pix++;
        }
    }
    return;
}

/* auxilliary routines that we'll be using in the main program */
void testpattern(int width, int height, int *red, int *green, int *blue);
void writeHTMLHeader(FILE *f);
void writeHTMLFooter(FILE *f);
void writeHTMLTitle(FILE *f, char *title);
void writeJSArray(FILE *f, char *arrname, int size, int *arr);
void writeHTMLImage(FILE *f, char *canvasname, int width, int height, int *r, int *g, int *b, int scale);


/* the main program.   Generate our input images, then process them */
int main(int argc, char **argv)
{
    const int lgncols = 256, lgnrows = 192;
    const int ncols = 24, nrows = 18;

    FILE *fp = fopen("example1.html","w");
    writeHTMLHeader(fp);

    /* display full-sized image we'll be using */
    writeHTMLTitle(fp, "Large Test image");
    int *lgr = (int *)malloc(lgncols*lgnrows*sizeof(int));
    int *lgg = (int *)malloc(lgncols*lgnrows*sizeof(int));
    int *lgb = (int *)malloc(lgncols*lgnrows*sizeof(int));
    testpattern(lgncols, lgnrows, lgr, lgg, lgb);
    writeHTMLImage(fp, "large", lgncols, lgnrows, lgr, lgg, lgb, 1);
    free(lgr); free(lgg); free(lgb);


    /* display scaled-down input image; plot it at a scale factor of 4 
     * so we can see it easier */
    writeHTMLTitle(fp, "Input Test image");
    int *r = (int *)malloc(ncols*nrows*sizeof(int));
    int *g = (int *)malloc(ncols*nrows*sizeof(int));
    int *b = (int *)malloc(ncols*nrows*sizeof(int));
    testpattern(ncols, nrows, r, g, b);
    writeHTMLImage(fp, "input", ncols, nrows, r, g, b, 4);


    /* convert it to greyscale using the CPU */
    writeHTMLTitle(fp, "CPU-greyscaled Test image");
    int *cpu_grey_r = (int *)malloc(ncols*nrows*sizeof(int));
    int *cpu_grey_g = (int *)malloc(ncols*nrows*sizeof(int));
    int *cpu_grey_b = (int *)malloc(ncols*nrows*sizeof(int));
    cpuGreyscale(ncols, nrows, r, g, b, cpu_grey_r, cpu_grey_g, cpu_grey_b);
    writeHTMLImage(fp, "grey_cpu", ncols, nrows, cpu_grey_r, cpu_grey_g, cpu_grey_b, 4);
    free(cpu_grey_r); free(cpu_grey_g); free(cpu_grey_b);


    /* convert it to greyscale using the GPU */
    writeHTMLTitle(fp, "GPU-greyscaled Test image");
    int *gpu_grey_r = (int *)malloc(ncols*nrows*sizeof(int));
    int *gpu_grey_g = (int *)malloc(ncols*nrows*sizeof(int));
    int *gpu_grey_b = (int *)malloc(ncols*nrows*sizeof(int));
    gpuGreyscale(ncols, nrows, r, g, b, gpu_grey_r, gpu_grey_g, gpu_grey_b);
    writeHTMLImage(fp, "grey_gpu", ncols, nrows, gpu_grey_r, gpu_grey_g, gpu_grey_b, 4);
    free(gpu_grey_r); free(gpu_grey_g); free(gpu_grey_b);

    writeHTMLFooter(fp);
    fclose(fp);

    free(r); free(g); free(b);
    return 0;
}

/* 
 * auxilliary routines -- 
 *   for generating the testpattern image,
 *   and for writing out the HTML file we'll be viewing as output.
 * You don't need to change anything below here.
 */

void testpattern(int width, int height, int *red, int *green, int *blue) {
    int i, r, c;

    int topbars[7][3] = {{0xc0,0xc0,0xc0}, {0xc0,0xc0,0x00}, {0x00,0xc0,0xc0}, {0x00,0xc0,0x00}, {0xc0,0x00,0xc0}, {0xc0,0x00,0x00}, {0x00,0x00,0xc0}};
    int midbars[7][3] = {{0x00,0x00,0xc0}, {0x13,0x13,0x13}, {0xc0,0x00,0xc0}, {0x13,0x13,0x13}, {0x00,0xc0,0xc0}, {0x13,0x13,0x13}, {0xc0,0xc0,0xc0}};
    int botbars[8][3] = {{0x00,0x21,0x4c}, {0xff,0xff,0xff}, {0x32,0x00,0x6a}, {0x13,0x13,0x13}, {0x09,0x09,0x09}, {0x13,0x13,0x13}, {0x1d,0x1d,0x1d}, {0x13,0x13,0x13}};

    int topheight = (int)round(height*334./504);
    int midheight = (int)round(height*43./504);

    int topmidwidths[7] = {94, 190, 287, 382, 480, 575, 671};
    for (i=0; i<7; i++)
        topmidwidths[i] = (int)round(topmidwidths[i]*width/672.);

    int botwidths[8] = {119, 239, 359, 479, 511, 543, 575, 671};
    for (i=0; i<8; i++)
        botwidths[i] = (int)round(botwidths[i]*width/672.);


    i = 0;
    for (r=0; r<height; r++) {
        if (r < topheight) {
            int bar = 0;
            for (c=0; c<width; c++) {
                if (c > topmidwidths[bar]) bar++;
                red[i] = topbars[bar][0]; green[i] = topbars[bar][1]; blue[i] = topbars[bar][2]; i++;
            }
        } else if (r < topheight + midheight) {
            int bar = 0;
            for (c=0; c<width; c++) {
                if (c > topmidwidths[bar]) bar++;
                red[i] = midbars[bar][0]; green[i] = midbars[bar][1]; blue[i] = midbars[bar][2]; i++;
            }
        } else {
            int bar = 0;
            for (c=0; c<width; c++) {
                if (c > botwidths[bar]) bar++;
                red[i] = botbars[bar][0]; green[i] = botbars[bar][1]; blue[i] = botbars[bar][2]; i++;
            }
        }
    }
    return;
}


void writeHTMLHeader(FILE *f) {
     fprintf(f, 
        "<head>"
        "    <title>Example 1</title>\n"
        "</head>\n"
        "<body>\n"
        "<script type=\"text/javascript\">\n"
        "function setPixel(imageData, x, y, r, g, b, a, scale) {\n"
        "    for (sx = 0; sx < scale; sx++) {\n"
        "        for (sy = 0; sy < scale; sy++) {\n"
        "           index = ((scale*x + sx) + (scale*y + sy)* imageData.width) * 4;\n"
        "           imageData.data[index+0] = Math.max(0,Math.min(255, r));\n"
        "           imageData.data[index+1] = Math.max(0,Math.min(255, g));\n"
        "           imageData.data[index+2] = Math.max(0,Math.min(255, b));\n"
        "           imageData.data[index+3] = Math.max(0,Math.min(255, a));\n"
        "        }\n"
        "    }\n"
        "}\n"
        "</script>\n"
        "    <h1>CPU and GPU Inversion of an image</h1>\n"
        "    <p>Below we'll see a full size test imaage; then a smaller one; then a CPU-inverted one; then a GPU-inverted one.</p>\n"
    );
}

void writeHTMLFooter(FILE *f) {
     fprintf(f, "</body>\n");
}

void writeHTMLTitle(FILE *f, char *title) {
     fprintf(f, "<h2>%s</h2>\n", title);
}

void writeJSArray(FILE *f, char *arrname, int size, int *arr) {
    int i;
    fprintf(f, "var %s = [", arrname);
    for (i=0; i<size-1; i++) 
        fprintf(f, "%3u, ", arr[i]);
    fprintf(f, "%3u];\n", arr[size-1]);
}

void writeHTMLImage(FILE *f, char *canvasname, int width, int height, int *r, int *g, int *b, int scale) {
    fprintf(f, 
        "<canvas id='%s' width='%d', height='%d'>\n"
        " Your browser does not support HTML5\n"
        "</canvas>\n", canvasname, width*scale, height*scale);
    fprintf(f,
        "<script type=\"text/javascript\">\n"
        "\n"
        "element = document.getElementById('%s');\n\n", canvasname);

    writeJSArray(f, "r", width*height, r);
    writeJSArray(f, "g", width*height, g);
    writeJSArray(f, "b", width*height, b);

    fprintf(f,"sc = %d;\n",scale);
    fprintf(f, 
        "c = element.getContext(\"2d\");\n"
        "\n"
        "// read the width and height of the canvas\n"
        "width = element.width;\n"
        "height = element.height;\n"
        "\n"
        "// create a new pixel array\n"
        "imageData = c.createImageData(width, height);\n"
        "\n"
        "pix = 0;\n"
        "for (y = 0; y < height/sc; y++) {\n"
        "    for (x = 0; x < width/sc; x++) {\n"
        "        setPixel(imageData, x, y, r[pix], g[pix], b[pix], 255, sc);\n"
        "        pix++;\n"
        "   }\n"
        "}\n"
        "\n"
        "c.putImageData(imageData, 0, 0);\n"
        "\n"
        "</script>\n");
}
