#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <cuda.h>

/* do a simple smoothing over the image:
 *
 *     +---+---+---+---+
 *     | 0 | 1 | 2 | 3 |
 *     +---+---+---+---+
 *     | 4 | 5 | 6 | 7 |
 *     +---+---+---+---+
 *     | 8 | 9 | 10| 11|
 *     +---+---+---+---+
 * 
 *   where the color in cell 5 will a weighted average of the
 *   values in cells 1, 4, 5, 6, and 9 - that is, cell 5 and it's 
 *   4 nearest neighbours.  In particular:
 *   
 *   red[5] = (red[1] + red[4] + red[5] + red[6] + red[9])/6;
 *  
 *   and so on for other color channels.
 *
 *   Note that the left and right neighbor of a pixel are pixel-1 and pixel+1;
 *   and the up and down neighbour of a pixel are pixel-ncols and pixle+ncols.
 *   Also note that we can't smooth the pixels at the ends of an image, so we just
 *   copy those values unchanged.
 */

__device__ __host__ int avg(const int *color, const int pix, const int nrows, const int ncols) {

    int avg = (               color[pix-ncols]
                + color[pix-1] + color[pix] + color[pix+1] +
                              color[pix+ncols]    ) / 5;

    return (int)avg;
}

void cpuSmooth(const int nrows, const int ncols, 
               const int *in_r, const int *in_g, const int *in_b,
               int *out_r, int *out_g, int *out_b) {

    int row, col;
    int pix = 0;
    for (row = 0; row<nrows; row++) {
        for (col = 0; col<ncols; col++) {
            if (row == 0 || row == nrows-1 || col == 0 || col == ncols-1) {
                out_r[pix] = in_r[pix];
                out_g[pix] = in_g[pix];
                out_b[pix] = in_b[pix];
            } else {
                out_r[pix] = avg(in_r, pix, nrows, ncols);
                out_g[pix] = avg(in_g, pix, nrows, ncols);
                out_b[pix] = avg(in_b, pix, nrows, ncols);
            }
            pix++;
        }
    }
    return;
}

/* handy macros for error testing with CUDA. */

#define CHK_CUDA(e) {if (e != cudaSuccess) {fprintf(stderr,"Error: %s\n", cudaGetErrorString(e)); exit(-1);}}
#define CHK_ERROR   {cudaThreadSynchronize(); cudaError_t e = cudaGetLastError(); if(e != cudaSuccess) {fprintf(stderr,"Error: %s\n", cudaGetErrorString(e)); exit(-1);}}


/* do a greyscale image by averaging the colors at a point */
__global__ void smoothKernel(const int nrows, const int ncols, 
                             const int *in_r_d, const int *in_g_d, const int *in_b_d,
                             int *out_r_d, int *out_g_d, int *out_b_d) {

        int pix = /* something else goes here */ 0;
        int row = pix/ncols;
        int col = (pix - row*ncols);

        if (pix < nrows*ncols) {
            /* do the appropriate work */
        }

        return;
}

/* launch the kernel on our input image */
void gpuSmooth(const int nrows, const int ncols, 
               const int *in_r, const int *in_g, const int *in_b,
               int *out_r, int *out_g, int *out_b, int blocksize) {

    int *in_r_d, *in_g_d, *in_b_d;
    int *out_r_d, *out_g_d, *out_b_d;

    /* cudaMalloc the device arrays*/

    /* cudaMemcpy the input data */

    /* calculate number of blocks we need -- round up */
    int nblocks = (nrows*ncols + blocksize - 1)/blocksize;
    smoothKernel<<<nblocks, blocksize>>>(nrows, ncols, in_r_d, in_g_d, in_b_d, out_r_d, out_g_d, out_b_d);

    CHK_ERROR ;

    /* cudaMemcpy the output data */

    /* cudaFree the device arrays */

    return;
}

/* auxilliary routines that we'll be using in the main program */
void testpattern(int width, int height, int *red, int *green, int *blue);
void writeHTMLHeader(FILE *f);
void writeHTMLFooter(FILE *f);
void writeHTMLTitle(FILE *f, char *title);
void writeJSArray(FILE *f, char *arrname, int size, int *arr);
void writeHTMLImage(FILE *f, char *canvasname, int width, int height, int *r, int *g, int *b, int scale);


/* the main program.   Generate our input images, then process them */
int main(int argc, char **argv)
{
    const int dimx = 256, dimy = 192;

    FILE *fp = fopen("example3.html","w");
    writeHTMLHeader(fp);

    /* display full-sized image we'll be using; scale up to better see blurring */
    writeHTMLTitle(fp, "Large Test image");
    int *r = (int *)malloc(dimx*dimy*sizeof(int));
    int *g = (int *)malloc(dimx*dimy*sizeof(int));
    int *b = (int *)malloc(dimx*dimy*sizeof(int));
    testpattern(dimx, dimy, r, g, b);
    writeHTMLImage(fp, "large", dimx, dimy, r, g, b, 2);

    /* smooth it with CPU */
    writeHTMLTitle(fp, "CPU-smoothed Test image");
    int *cpu_smooth_r = (int *)malloc(dimx*dimy*sizeof(int));
    int *cpu_smooth_g = (int *)malloc(dimx*dimy*sizeof(int));
    int *cpu_smooth_b = (int *)malloc(dimx*dimy*sizeof(int));
    cpuSmooth(dimy, dimx, r, g, b, cpu_smooth_r, cpu_smooth_g, cpu_smooth_b);
    writeHTMLImage(fp, "smooth_cpu", dimx, dimy, cpu_smooth_r, cpu_smooth_g, cpu_smooth_b, 2);
    free(cpu_smooth_r); free(cpu_smooth_g); free(cpu_smooth_b);


    /* smooth it with GPU */
    writeHTMLTitle(fp, "GPU-smoothed Test image");
    int *gpu_smooth_r = (int *)calloc(dimx*dimy, sizeof(int));
    int *gpu_smooth_g = (int *)calloc(dimx*dimy, sizeof(int));
    int *gpu_smooth_b = (int *)calloc(dimx*dimy, sizeof(int));
    gpuSmooth(dimy, dimx, r, g, b, gpu_smooth_r, gpu_smooth_g, gpu_smooth_b, 256);
    writeHTMLImage(fp, "smooth_gpu", dimx, dimy, gpu_smooth_r, gpu_smooth_g, gpu_smooth_b, 2);
    free(gpu_smooth_r); free(gpu_smooth_g); free(gpu_smooth_b);

    writeHTMLFooter(fp);
    fclose(fp);

    free(r); free(g); free(b);
    return 0;
}

/* 
 * auxilliary routines -- 
 *   for generating the testpattern image,
 *   and for writing out the HTML file we'll be viewing as output.
 * You don't need to change anything below here.
 */

void testpattern(int width, int height, int *red, int *green, int *blue) {
    int i, r, c;

    int topbars[7][3] = {{0xc0,0xc0,0xc0}, {0xc0,0xc0,0x00}, {0x00,0xc0,0xc0}, {0x00,0xc0,0x00}, {0xc0,0x00,0xc0}, {0xc0,0x00,0x00}, {0x00,0x00,0xc0}};
    int midbars[7][3] = {{0x00,0x00,0xc0}, {0x13,0x13,0x13}, {0xc0,0x00,0xc0}, {0x13,0x13,0x13}, {0x00,0xc0,0xc0}, {0x13,0x13,0x13}, {0xc0,0xc0,0xc0}};
    int botbars[8][3] = {{0x00,0x21,0x4c}, {0xff,0xff,0xff}, {0x32,0x00,0x6a}, {0x13,0x13,0x13}, {0x09,0x09,0x09}, {0x13,0x13,0x13}, {0x1d,0x1d,0x1d}, {0x13,0x13,0x13}};

    int topheight = (int)round(height*334./504);
    int midheight = (int)round(height*43./504);

    int topmidwidths[7] = {94, 190, 287, 382, 480, 575, 671};
    for (i=0; i<7; i++)
        topmidwidths[i] = (int)round(topmidwidths[i]*width/672.);

    int botwidths[8] = {119, 239, 359, 479, 511, 543, 575, 671};
    for (i=0; i<8; i++)
        botwidths[i] = (int)round(botwidths[i]*width/672.);


    i = 0;
    for (r=0; r<height; r++) {
        if (r < topheight) {
            int bar = 0;
            for (c=0; c<width; c++) {
                if (c > topmidwidths[bar]) bar++;
                red[i] = topbars[bar][0]; green[i] = topbars[bar][1]; blue[i] = topbars[bar][2]; i++;
            }
        } else if (r < topheight + midheight) {
            int bar = 0;
            for (c=0; c<width; c++) {
                if (c > topmidwidths[bar]) bar++;
                red[i] = midbars[bar][0]; green[i] = midbars[bar][1]; blue[i] = midbars[bar][2]; i++;
            }
        } else {
            int bar = 0;
            for (c=0; c<width; c++) {
                if (c > botwidths[bar]) bar++;
                red[i] = botbars[bar][0]; green[i] = botbars[bar][1]; blue[i] = botbars[bar][2]; i++;
            }
        }
    }
    return;
}


void writeHTMLHeader(FILE *f) {
     fprintf(f, 
        "<head>"
        "    <title>Example 1</title>\n"
        "</head>\n"
        "<body>\n"
        "<script type=\"text/javascript\">\n"
        "function setPixel(imageData, x, y, r, g, b, a, scale) {\n"
        "    for (sx = 0; sx < scale; sx++) {\n"
        "        for (sy = 0; sy < scale; sy++) {\n"
        "           index = ((scale*x + sx) + (scale*y + sy)* imageData.width) * 4;\n"
        "           imageData.data[index+0] = Math.max(0,Math.min(255, r));\n"
        "           imageData.data[index+1] = Math.max(0,Math.min(255, g));\n"
        "           imageData.data[index+2] = Math.max(0,Math.min(255, b));\n"
        "           imageData.data[index+3] = Math.max(0,Math.min(255, a));\n"
        "        }\n"
        "    }\n"
        "}\n"
        "</script>\n"
        "    <h1>CPU and GPU Inversion of an image</h1>\n"
        "    <p>Below we'll see a test image; then a CPU-smoothed one; then a GPU-smoothed one.</p>\n"
    );
}

void writeHTMLFooter(FILE *f) {
     fprintf(f, "</body>\n");
}

void writeHTMLTitle(FILE *f, char *title) {
     fprintf(f, "<h2>%s</h2>\n", title);
}

void writeJSArray(FILE *f, char *arrname, int size, int *arr) {
    int i;
    fprintf(f, "var %s = [", arrname);
    for (i=0; i<size-1; i++) 
        fprintf(f, "%3u, ", arr[i]);
    fprintf(f, "%3u];\n", arr[size-1]);
}

void writeHTMLImage(FILE *f, char *canvasname, int width, int height, int *r, int *g, int *b, int scale) {
    fprintf(f, 
        "<canvas id='%s' width='%d', height='%d'>\n"
        " Your browser does not support HTML5\n"
        "</canvas>\n", canvasname, width*scale, height*scale);
    fprintf(f,
        "<script type=\"text/javascript\">\n"
        "\n"
        "element = document.getElementById('%s');\n\n", canvasname);

    writeJSArray(f, "r", width*height, r);
    writeJSArray(f, "g", width*height, g);
    writeJSArray(f, "b", width*height, b);

    fprintf(f,"sc = %d;\n",scale);
    fprintf(f, 
        "c = element.getContext(\"2d\");\n"
        "\n"
        "// read the width and height of the canvas\n"
        "width = element.width;\n"
        "height = element.height;\n"
        "\n"
        "// create a new pixel array\n"
        "imageData = c.createImageData(width, height);\n"
        "\n"
        "pix = 0;\n"
        "for (y = 0; y < height/sc; y++) {\n"
        "    for (x = 0; x < width/sc; x++) {\n"
        "        setPixel(imageData, x, y, r[pix], g[pix], b[pix], 255, sc);\n"
        "        pix++;\n"
        "   }\n"
        "}\n"
        "\n"
        "c.putImageData(imageData, 0, 0);\n"
        "\n"
        "</script>\n");
}
