import numpy
import matplotlib
import matplotlib.pylab as plt

import pycuda.driver as cuda
import pycuda.autoinit
from pycuda.compiler import SourceModule

def testpattern(width, height):
    topbars = [[0xc0,0xc0,0xc0], [0xc0,0xc0,0x00], [0x00,0xc0,0xc0],
               [0x00,0xc0,0x00], [0xc0,0x00,0xc0], [0xc0,0x00,0x00],
               [0x00,0x00,0xc0]]
    midbars = [[0x00,0x00,0xc0], [0x13,0x13,0x13], [0xc0,0x00,0xc0],
               [0x13,0x13,0x13], [0x00,0xc0,0xc0], [0x13,0x13,0x13],
               [0xc0,0xc0,0xc0]]
    botbars = [[0x00,0x21,0x4c], [0xff,0xff,0xff], [0x32,0x00,0x6a],
               [0x13,0x13,0x13], [0x09,0x09,0x09], [0x13,0x13,0x13],
               [0x1d,0x1d,0x1d], [0x13,0x13,0x13]]

    topheight = int(height*334./504.)
    midheight = int(height*43./504.) + topheight

    topmidwidths = numpy.array([94, 190, 287, 382, 480, 575, 671])/672.*width
    botwidths = numpy.array([119, 239, 359, 479, 511, 543, 575, 671.])/672.*width

    r = numpy.zeros((height,width),dtype=numpy.int32)
    g = numpy.zeros_like(r)
    b = numpy.zeros_like(r)

    lastwidth = 0
    for wide, color in zip(topmidwidths, topbars):
        r[0:topheight, lastwidth:wide] = color[0]
        g[0:topheight, lastwidth:wide] = color[1]
        b[0:topheight, lastwidth:wide] = color[2]
        lastwidth = wide

    lastwidth = 0
    for wide, color in zip(topmidwidths, midbars):
        r[topheight:midheight, lastwidth:wide] = color[0]
        g[topheight:midheight, lastwidth:wide] = color[1]
        b[topheight:midheight, lastwidth:wide] = color[2]
        lastwidth = wide

    lastwidth = 0
    for wide, color in zip(botwidths, botbars):
        r[midheight:, lastwidth:wide] = color[0]
        g[midheight:, lastwidth:wide] = color[1]
        b[midheight:, lastwidth:wide] = color[2]
        lastwidth = wide
        
    r.shape=(width*height)
    g.shape=(width*height)
    b.shape=(width*height)

    return r, g, b

def plotimage(width, height, r, g, b, title):
    fig = plt.figure()
    img = numpy.array([[r[i]/256., g[i]/256., b[i]/256.] for i in xrange(len(r))])
    img.shape=(height, width, 3)
    plt.imshow(img)
    fig.suptitle(title)
    plt.show()

def cpuGreyscale(nrows, ncols, in_r, in_g, in_b, out_r, out_g, out_b):
    for i in xrange(nrows*ncols):
        avg = int( (in_r[i] + in_g[i] + in_b[i])/3. )
        out_r[i] = avg
        out_g[i] = avg
        out_b[i] = avg

# Main program starts here

width = 32
height = width*3/4
r, g, b = testpattern(width, height)
plotimage(width, height, r, g, b, "Original")

# Greyscale the image on the CPU

cpu_r = numpy.zeros_like(r)
cpu_g = numpy.zeros_like(g)
cpu_b = numpy.zeros_like(b)

cpuGreyscale(height, width, r, g, b, cpu_r, cpu_g, cpu_b)
plotimage(width, height, cpu_r, cpu_g, cpu_b, "CPU Greyscale")

# Greyscale the image on the GPU
# here's the CUDA kernel:

cudaSource = SourceModule("""__global__ void greyscaleKernel(const int nrows, const int ncols,
                             const int *in_r_d, 
                             const int *in_g_d, 
                             const int *in_b_d,
                             int *out_r_d, int *out_g_d, int *out_b_d) {

        int i = threadIdx.x;

        if (i < nrows*ncols) {
            int avg = (in_r_d[i] + in_g_d[i] + in_b_d[i])/3 ;
            out_r_d[i] = avg;
            out_g_d[i] = avg;
            out_b_d[i] = avg;
        }

        return;
}""")

# build CUDA kernel
gpuGreyscale = cudaSource.get_function("greyscaleKernel")

# allocate memory, copy input data over
in_r_gpu = cuda.mem_alloc( r.nbytes )
in_g_gpu = cuda.mem_alloc( g.nbytes )
in_b_gpu = cuda.mem_alloc( b.nbytes )
cuda.memcpy_htod( in_r_gpu, r )
cuda.memcpy_htod( in_g_gpu, g )
cuda.memcpy_htod( in_b_gpu, b )
# allocate memory for outputs
out_r_gpu = cuda.mem_alloc( r.nbytes )
out_g_gpu = cuda.mem_alloc( g.nbytes )
out_b_gpu = cuda.mem_alloc( b.nbytes )

# run kernel
gpuGreyscale(numpy.int32(height), numpy.int32(width), in_r_gpu, in_g_gpu, in_b_gpu, out_r_gpu, out_g_gpu, out_b_gpu, grid=(1,1), block=(width*height,1,1))

# copy data back
gpu_r = numpy.zeros_like(r)
gpu_g = numpy.zeros_like(g)
gpu_b = numpy.zeros_like(b)

cuda.memcpy_dtoh(gpu_r, out_r_gpu)
cuda.memcpy_dtoh(gpu_g, out_g_gpu)
cuda.memcpy_dtoh(gpu_b, out_b_gpu)

# and plot it
plotimage(width, height, gpu_r, gpu_g, gpu_b, "GPU Greyscale")

