import numpy
import matplotlib
import matplotlib.pylab as plt

import pycuda.driver as cuda
import pycuda.autoinit
from pycuda.compiler import SourceModule

def testpattern(width, height):
    topbars = [[0xc0,0xc0,0xc0], [0xc0,0xc0,0x00], [0x00,0xc0,0xc0],
               [0x00,0xc0,0x00], [0xc0,0x00,0xc0], [0xc0,0x00,0x00],
               [0x00,0x00,0xc0]]
    midbars = [[0x00,0x00,0xc0], [0x13,0x13,0x13], [0xc0,0x00,0xc0],
               [0x13,0x13,0x13], [0x00,0xc0,0xc0], [0x13,0x13,0x13],
               [0xc0,0xc0,0xc0]]
    botbars = [[0x00,0x21,0x4c], [0xff,0xff,0xff], [0x32,0x00,0x6a],
               [0x13,0x13,0x13], [0x09,0x09,0x09], [0x13,0x13,0x13],
               [0x1d,0x1d,0x1d], [0x13,0x13,0x13]]

    topheight = int(height*334./504.)
    midheight = int(height*43./504.) + topheight

    topmidwidths = numpy.array([94, 190, 287, 382, 480, 575, 671])/672.*width
    botwidths = numpy.array([119, 239, 359, 479, 511, 543, 575, 671.])/672.*width

    r = numpy.zeros((height,width),dtype=numpy.int32)
    g = numpy.zeros_like(r)
    b = numpy.zeros_like(r)

    lastwidth = 0
    for wide, color in zip(topmidwidths, topbars):
        r[0:topheight, lastwidth:wide] = color[0]
        g[0:topheight, lastwidth:wide] = color[1]
        b[0:topheight, lastwidth:wide] = color[2]
        lastwidth = wide

    lastwidth = 0
    for wide, color in zip(topmidwidths, midbars):
        r[topheight:midheight, lastwidth:wide] = color[0]
        g[topheight:midheight, lastwidth:wide] = color[1]
        b[topheight:midheight, lastwidth:wide] = color[2]
        lastwidth = wide

    lastwidth = 0
    for wide, color in zip(botwidths, botbars):
        r[midheight:, lastwidth:wide] = color[0]
        g[midheight:, lastwidth:wide] = color[1]
        b[midheight:, lastwidth:wide] = color[2]
        lastwidth = wide
        
    r.shape=(width*height)
    g.shape=(width*height)
    b.shape=(width*height)

    return r, g, b

def plotimage(width, height, r, g, b, title):
    fig = plt.figure()
    img = numpy.array([[r[i]/256., g[i]/256., b[i]/256.] for i in xrange(len(r))])
    img.shape=(height, width, 3)
    plt.imshow(img)
    fig.suptitle(title)
    plt.show()

# do a simple smoothing over the image:
#
#     +---+---+---+---+
#     | 0 | 1 | 2 | 3 |
#     +---+---+---+---+
#     | 4 | 5 | 6 | 7 |
#     +---+---+---+---+
#     | 8 | 9 | 10| 11|
#     +---+---+---+---+
#
#   where the color in cell 5 will a weighted average of the
#   values in cells 1, 4, 5, 6, and 9 - that is, cell 5 and it's
#   4 nearest neighbours.  In particular:
#
#   red[5] = (red[1] + red[4] + red[5] + red[6] + red[9])/6;
#
#   and so on for other color channels.
#
#   Note that the left and right neighbor of a pixel are pixel-1 and pixel+1;
#   and the up and down neighbour of a pixel are pixel-ncols and pixle+ncols.
#   Also note that we can't smooth the pixels at the ends of an image, so we just
#   copy those values unchanged.

def avg(color, pix, nrows, ncols):
    avgpix  = (               color[pix-ncols]
                + color[pix-1] + color[pix] + color[pix+1] +
                              color[pix+ncols]    ) / 5
    return avgpix

def cpuSmooth(nrows, ncols, in_r, in_g, in_b, out_r, out_g, out_b):
    for pix in xrange(nrows*ncols):
        row = pix / ncols
        col = pix % ncols
        if (row == 0 or row == nrows-1 or col == 0 or col == ncols-1):
                out_r[pix] = in_r[pix]
                out_g[pix] = in_g[pix]
                out_b[pix] = in_b[pix]
        else:
                out_r[pix] = avg(in_r, pix, nrows, ncols)
                out_g[pix] = avg(in_g, pix, nrows, ncols)
                out_b[pix] = avg(in_b, pix, nrows, ncols)

# Main program starts here

width = 256
height = width*3/4
r, g, b = testpattern(width, height)
plotimage(width, height, r, g, b, "Original")

# Greyscale the image on the CPU

cpu_r = numpy.zeros_like(r)
cpu_g = numpy.zeros_like(g)
cpu_b = numpy.zeros_like(b)

cpuSmooth(height, width, r, g, b, cpu_r, cpu_g, cpu_b)
plotimage(width, height, cpu_r, cpu_g, cpu_b, "CPU Smoothed")

# Greyscale the image on the GPU
# here's the CUDA kernel:

cudaSource = SourceModule("""
__global__ void smoothKernel(const int nrows, const int ncols,
                             const int *in_r_d, const int *in_g_d, const int *in_b_d,
                             int *out_r_d, int *out_g_d, int *out_b_d) {

        int pix = /* something else goes here */ 0;
        int row = pix/ncols;
        int col = (pix - row*ncols);

        if (pix < nrows*ncols) {
            /* do the appropriate work */
        }

        return;
}""")

# build CUDA kernel
gpuSmooth = cudaSource.get_function("smoothKernel")

# allocate memory, copy input data over
# cuda.mem_alloc( ... something? ... )
# cuda.memcpy_... something?
# allocate memory for outputs
# cuda.mem_alloc( ... something? ... )

# run kernel
blocksize = 128
gridsize  = (width*height + blocksize - 1)/blocksize
gpuSmooth(numpy.int32(height), numpy.int32(width), in_r_gpu, in_g_gpu, in_b_gpu, 
              out_r_gpu, out_g_gpu, out_b_gpu, grid=(gridsize,1), block=(blocksize,1,1))

gpu_r = numpy.zeros_like(r)
gpu_g = numpy.zeros_like(g)
gpu_b = numpy.zeros_like(b)
# copy data back
# cuda.memcpy_... something?

# and plot it
plotimage(width, height, gpu_r, gpu_g, gpu_b, "GPU Smoothed")

